#include "llvm/IRReader/IRReader.h"
#include "llvm/IR/Module.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/Bitcode/BitstreamWriter.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/PassManager.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/IPO.h"
#include "llvm/Transforms/Vectorize.h"
#include "llvm/Support/TargetRegistry.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetLowering.h"
#include "llvm/Target/TargetLibraryInfo.h"
#include "llvm/Support/ToolOutputFile.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/Support/FormattedStream.h"
#include "llvm/IR/DataLayout.h"
#include "llvm/Support/SourceMgr.h"
#include "llvm/Assembly/Parser.h"
#include "llvm/IR/TypeBuilder.h"

#include "llvm/ADT/ArrayRef.h"
#include "llvm/ADT/StringSet.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/Attributes.h"
#include "llvm/IR/DerivedTypes.h"
#include "llvm/IR/Attributes.h"

#include "llvm/Support/raw_os_ostream.h"

#include "llvm/Support/DataStream.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/Support/Program.h"
#include "llvm/Support/system_error.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Host.h"

#include "llvm/Linker.h"
#include "llvm/Assembly/PrintModulePass.h"

#include "llvm/ExecutionEngine/ObjectBuffer.h"
#include "llvm/IR/GlobalVariable.h"

#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/MCJIT.h"
#include "llvm/Analysis/Verifier.h"
#include "llvm/Transforms/Utils/Cloning.h"

#include <string>
#include <vector>
#include <iostream>

using namespace llvm;
using namespace std;



  llvm::IRBuilder<> *builder;
  llvm::BasicBlock  *entry;
  llvm::Function    *mainFunc;
  llvm::Function    *mainFunc_extern;
  llvm::Module      *Mod;
  llvm::ExecutionEngine *TheExecutionEngine;

  llvm::TargetMachine *targetMachine;

  llvm::FunctionPassManager *TheFPM;

  void * fptr_mainFunc_extern;




int main(int argc, char **argv) {

    llvm::InitializeNativeTarget();

    InitializeNativeTargetAsmPrinter(); // MCJIT
    InitializeNativeTargetAsmParser(); // MCJIT

    string fileName = argv[1];

    SMDiagnostic Err;
    Mod = llvm::ParseIRFile(fileName.c_str(), Err, llvm::getGlobalContext());

    std::cout << "File parsed\n";

    llvm::outs() << "detected process triple = " << llvm::sys::getProcessTriple() << "\n";

    Mod->setTargetTriple(llvm::sys::getProcessTriple());

    string error;
    llvm::EngineBuilder engineBuilder(Mod);

    llvm::outs() << "detected HostCPUName = " << llvm::sys::getHostCPUName() << "\n";

    engineBuilder.setMCPU(llvm::sys::getHostCPUName());

    engineBuilder.setEngineKind(llvm::EngineKind::JIT);
    engineBuilder.setOptLevel(llvm::CodeGenOpt::Aggressive);
    engineBuilder.setErrorStr(&error);

    TheExecutionEngine = engineBuilder.setUseMCJIT(true).create(); // MCJIT

    assert(TheExecutionEngine && "failed to create LLVM ExecutionEngine with error");

    targetMachine = engineBuilder.selectTarget();

    mainFunc = Mod->getFunction("func");

    assert( mainFunc && "function func not found" );

    std::cout << "Verifying function\n";
    llvm::verifyFunction(*mainFunc);

    std::cout << "IR (before passes)\n";
    mainFunc->dump();


#if 1
    // Write asm file
    llvm::Module      *Mod_clone;
    Mod_clone = CloneModule(Mod);
    const string extension = fileName.substr(fileName.find_last_of(".") + 1);
    string errorInfo;
    string fileName_o = fileName.substr(0,fileName.find_last_of(".")) + ".s";
    std::cout << "writing " << fileName_o << "\n";
    tool_output_file output( fileName_o.c_str() , errorInfo , sys::fs::F_None );
    PassManager pm;
    formatted_raw_ostream fos(output.os());
    targetMachine->addPassesToEmitFile(pm, fos, TargetMachine::CGFT_AssemblyFile ); //;
    output.keep();
    pm.run(*Mod_clone);
    // End 
#endif
    // TargetMachine::CGFT_AssemblyFile
    // TargetMachine::CGFT_ObjectFile


    TheExecutionEngine->finalizeObject();  // MCJIT

    void *fptr = TheExecutionEngine->getPointerToFunction( mainFunc );

    // Now try and call the function

    void (*FP)(long int , long int , long int , bool , long int , double* , double*, double* ) = (void (*)(long int , long int , long int , bool , long int , double* , double*, double*))(intptr_t)fptr;
  
    const size_t s = 16*16*16*16;
    double *a = new double[s*2*3*3];
    double *b = new double[s*2*3*3];
    double *c = new double[s*2*3*3];

    for ( int i = 0 ; i < s*2*3*3 ; ++i ) {
      a[i] = -1.0;
      b[i] = 0.05;
      c[i] = 0.05;
    }

    std::cout << "before calling JIT'ed function\n";
    FP(0,s,0,true,0,a,b,c);
    std::cout << "after calling JIT'ed function\n";

    for ( int i = 0 ; i < 10 ; ++i )
      std::cout << a[i] << " ";
    std::cout << "\n";

  return 0;
}
